#include "GlFrameBuffer.h"
#include "GlDownloadPixelBuffer.h"
#include "GlTexture.h"
#include "OpenGl.h"
#include "Buffer.h"

#define GL_BUFFER_OFFSET( n ) ((char* )NULL + ( n ) )

namespace ProceduralTextures
{
	namespace gl
	{
		FrameBuffer::FrameBuffer( std::shared_ptr< OpenGl > p_pOpenGl )
			: Object( p_pOpenGl,
					  std::bind( &OpenGl::GenFramebuffers, p_pOpenGl, std::placeholders::_1, std::placeholders::_2 ),
					  std::bind( &OpenGl::DeleteFramebuffers, p_pOpenGl, std::placeholders::_1, std::placeholders::_2 ),
					  std::bind( &OpenGl::IsFramebuffer, p_pOpenGl, std::placeholders::_1 ),
					  std::bind( &OpenGl::BindFramebuffer, p_pOpenGl, GL_FRAMEBUFFER, std::placeholders::_1 )
					)
			, m_bInitialised( false )
		{
		}

		FrameBuffer::~FrameBuffer()
		{
		}

		bool FrameBuffer::Initialise( Size const & p_size )
		{
			if ( !m_bInitialised && !m_mapAttachments.empty() )
			{
				m_sizeImage = p_size;
				Bind();

				for ( auto l_pair : m_mapAttachments )
				{
					GetOpenGl()->FramebufferTexture2D( GL_FRAMEBUFFER, l_pair.first, GL_TEXTURE_2D, l_pair.second.lock()->GetGlName(), 0 );
				}

				GLenum l_eStatus = GetOpenGl()->CheckFramebufferStatus( GL_FRAMEBUFFER );
				Unbind();
				DoInitialisePbos();
				m_bInitialised = ( l_eStatus == GL_FRAMEBUFFER_COMPLETE );
			}

			return m_bInitialised;
		}

		void FrameBuffer::Cleanup()
		{
			m_bInitialised = false;
			DoCleanupPbos();
		}

		bool FrameBuffer::ActivateTexture( GLenum p_attachment, unsigned int p_target )
		{
			auto l_it = m_mapAttachments.find( p_attachment );
			bool l_bReturn = l_it != m_mapAttachments.end();

			if ( l_bReturn )
			{
				l_bReturn = l_it->second.lock()->Activate( p_target );
			}

			return l_bReturn;
		}

		void FrameBuffer::DeactivateTexture( GLenum p_attachment, unsigned int p_target )
		{
			auto l_it = m_mapAttachments.find( p_attachment );

			if ( l_it != m_mapAttachments.end() )
			{
				l_it->second.lock()->Deactivate( p_target );
			}
		}

		void FrameBuffer::AttachTexture( GLenum p_attachment, std::shared_ptr< Texture > p_pTexture )
		{
			m_mapAttachments[p_attachment] = p_pTexture;
		}

		std::shared_ptr< Texture > FrameBuffer::GetTexture( GLenum p_attachment )
		{
			std::shared_ptr< Texture > l_pReturn;
			auto l_it = m_mapAttachments.find( p_attachment );

			if ( l_it != m_mapAttachments.end() )
			{
				l_pReturn = l_it->second.lock();
			}

			return l_pReturn;
		}

		bool FrameBuffer::DownloadSync( GLenum p_attachment, PixelBuffer & p_buffer )
		{
			bool l_bReturn = false;

			if ( GetOpenGl()->ReadBuffer( p_attachment ) )
			{
				l_bReturn = GetOpenGl()->ReadPixels( 0, 0, m_sizeImage.x(), m_sizeImage.y(), GL_RGBA, GL_UNSIGNED_BYTE, p_buffer.Ptr() );
			}

			return l_bReturn;
		}

		bool FrameBuffer::DownloadAsync( GLenum p_attachment, PixelBuffer & p_buffer )
		{
			bool l_bReturn = false;

			if ( GetOpenGl()->ReadBuffer( p_attachment ) )
			{
				m_iCurrentDlPbo = ( m_iCurrentDlPbo + 1 ) % 2;
				int l_nextDlPbo = ( m_iCurrentDlPbo + 1 ) % 2;
				std::unique_ptr< DownloadPixelBuffer > & pBufferOut = m_pDownloadBuffers[m_iCurrentDlPbo];
				std::unique_ptr< DownloadPixelBuffer > & pBufferIn = m_pDownloadBuffers[l_nextDlPbo];

				if ( pBufferIn && pBufferOut )
				{
					if ( pBufferOut->Activate() )
					{
						GetOpenGl()->ReadPixels( 0, 0, m_sizeImage.x(), m_sizeImage.y(), GL_RGBA, GL_UNSIGNED_BYTE, 0 );

						if ( pBufferIn->Activate() )
						{
							void * pData = pBufferIn->Lock( GL_READ_ONLY );

							if ( pData )
							{
								memcpy( p_buffer.Ptr(), pData, p_buffer.GetSize() );
								pBufferIn->Unlock();
								l_bReturn = true;
							}

							pBufferIn->Deactivate();
						}

						pBufferOut->Deactivate();
					}
				}
			}

			return l_bReturn;
		}

		void FrameBuffer::Resize( Size const & p_size )
		{
			m_size = p_size;
			m_bInitialised = false;
		}

		void FrameBuffer::DoCleanupPbos()
		{
			if ( m_pDownloadBuffers[0] )
			{
				m_pDownloadBuffers[0]->Destroy();
				m_pDownloadBuffers[0].reset();
			}

			if ( m_pDownloadBuffers[1] )
			{
				m_pDownloadBuffers[1]->Destroy();
				m_pDownloadBuffers[1].reset();
			}
		}

		void FrameBuffer::DoInitialisePbos()
		{
			if ( !m_pDownloadBuffers[0] )
			{
				m_pDownloadBuffers[0] = std::make_unique< DownloadPixelBuffer >( GetOpenGl() );
				m_pDownloadBuffers[0]->Create();
				m_pDownloadBuffers[0]->Initialise( m_sizeImage.x() * m_sizeImage.y() * 4 );
			}

			if ( !m_pDownloadBuffers[0] )
			{
				m_pDownloadBuffers[1] = std::make_unique< DownloadPixelBuffer >( GetOpenGl() );
				m_pDownloadBuffers[1]->Create();
				m_pDownloadBuffers[1]->Initialise( m_sizeImage.x() * m_sizeImage.y() * 4 );
			}
		}
	}
}
